"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_general_1 = require("@energyweb/utils-general");
const __1 = require("..");
const getAccountFromConfiguration = (configuration) => ({
    from: configuration.blockchainProperties.activeUser.address,
    privateKey: configuration.blockchainProperties.activeUser.privateKey
});
exports.getAllCertificateEvents = async (certId, configuration) => {
    const registry = configuration.blockchainProperties.certificateLogicInstance;
    const allEvents = await registry.getAllEvents({
        topics: [
            null,
            configuration.blockchainProperties.web3.utils.padLeft(configuration.blockchainProperties.web3.utils.fromDecimal(certId), 64, '0')
        ],
        fromBlock: 0,
        toBlock: 'latest'
    });
    const returnEvents = [];
    for (const fullEvent of allEvents) {
        // we have to remove some false positives due to ERC721 interface
        if (fullEvent.event === 'Transfer') {
            if (fullEvent.returnValues.tokenId === `${certId}`) {
                returnEvents.push(fullEvent);
            }
        }
        else {
            returnEvents.push(fullEvent);
        }
    }
    // we also have to search
    if (certId !== 0) {
        const transferEvents = await registry.getAllTransferSingleEvents({
            topics: [
                '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef',
                null,
                null,
                configuration.blockchainProperties.web3.utils.padLeft(configuration.blockchainProperties.web3.utils.fromDecimal(certId), 64, '0')
            ],
            fromBlock: 0,
            toBlock: 'latest'
        });
        for (const transferEvent of transferEvents) {
            returnEvents.push(transferEvent);
        }
    }
    return returnEvents;
};
class Entity extends utils_general_1.BlockchainDataModelEntity.Entity {
    constructor(id, configuration, isPrivate = false) {
        super(id, configuration);
        this.isPrivate = isPrivate;
        this.initialized = false;
    }
    async sync() {
        if (this.id === null) {
            return this;
        }
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        const publicIssuer = this.configuration.blockchainProperties.issuerLogicInstance.public;
        const cert = await registry.getCertificate(Number(this.id));
        this.data = cert.data;
        const decodedData = await publicIssuer.decodeIssue(this.data);
        this.generationStartTime = Number(decodedData['0']);
        this.generationEndTime = Number(decodedData['1']);
        this.deviceId = decodedData['2'];
        this.issuer = cert.issuer;
        const allIssuanceEvents = await registry.getAllIssuanceSingleEvents({
            filter: { _id: Number(this.id) }
        });
        const creationBlock = await this.configuration.blockchainProperties.web3.eth.getBlock(allIssuanceEvents[0].blockNumber);
        this.creationTime = Number(creationBlock.timestamp);
        this.initialized = true;
        if (this.configuration.logger) {
            this.configuration.logger.verbose(`Certificate ${this.id} synced`);
        }
        return this;
    }
    async claim(amount) {
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        const owner = this.configuration.blockchainProperties.activeUser.address;
        const { randomHex, hexToBytes } = this.configuration.blockchainProperties.web3.utils;
        // TO-DO: replace with proper claim data
        const claimData = hexToBytes(randomHex(32));
        return registry.safeTransferAndClaimFrom(owner, owner, parseInt(this.id, 10), amount, this.data, claimData, getAccountFromConfiguration(this.configuration));
    }
    async transfer(to, amount) {
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        const from = this.configuration.blockchainProperties.activeUser.address;
        return registry.safeTransferFrom(from, to, parseInt(this.id, 10), amount, this.data, getAccountFromConfiguration(this.configuration));
    }
    async revoke() {
        const publicIssuer = this.configuration.blockchainProperties.issuerLogicInstance.public;
        return publicIssuer.revokeCertificate(Number(this.id), getAccountFromConfiguration(this.configuration));
    }
    async getAllCertificateEvents() {
        return exports.getAllCertificateEvents(parseInt(this.id, 10), this.configuration);
    }
    async isOwned() {
        if (this.isPrivate) {
            throw new Error('Unable to fetch owner for private certificates.');
        }
        const ownedVolume = await this.ownedVolume();
        return ownedVolume > 0;
    }
    async ownedVolume() {
        if (this.isPrivate) {
            throw new Error('Unable to fetch volumes for private certificates.');
        }
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        const address = this.configuration.blockchainProperties.activeUser.address;
        const balance = await registry.balanceOf(address, Number(this.id), { from: address });
        return Number(balance);
    }
    async isClaimed() {
        const claimedVolume = await this.claimedVolume();
        return claimedVolume > 0;
    }
    async claimedVolume() {
        const registry = this.configuration.blockchainProperties.certificateLogicInstance;
        const address = this.configuration.blockchainProperties.activeUser.address;
        const balance = await registry.claimedBalanceOf(address, Number(this.id), getAccountFromConfiguration(this.configuration));
        return Number(balance);
    }
}
exports.Entity = Entity;
exports.createCertificate = async (to, value, fromTime, toTime, deviceId, configuration, isVolumePrivate = false) => {
    const certificate = new Entity(null, configuration, isVolumePrivate);
    const registry = configuration.blockchainProperties.certificateLogicInstance;
    const getIdFromLogs = (logs) => configuration.blockchainProperties.web3.utils
        .hexToNumber(logs[0].topics[2])
        .toString();
    if (isVolumePrivate) {
        const privateIssuer = configuration.blockchainProperties.issuerLogicInstance.private;
        const data = await privateIssuer.encodeIssue(fromTime, toTime, deviceId);
        const { logs } = await registry.issue(to, [], __1.PRIVATE_CERTIFICATE_TOPIC, value, data, getAccountFromConfiguration(configuration));
        certificate.id = getIdFromLogs(logs);
        const commitment = { volume: value };
        const { rootHash } = certificate.prepareEntityCreation(commitment, __1.CommitmentSchema);
        await privateIssuer.updateCommitment(Number(certificate.id), configuration.blockchainProperties.web3.utils.hexToBytes('0x0'), rootHash);
    }
    else {
        const publicIssuer = configuration.blockchainProperties.issuerLogicInstance.public;
        const data = await publicIssuer.encodeIssue(fromTime, toTime, deviceId);
        const { logs } = await publicIssuer.issue(to, value, data, getAccountFromConfiguration(configuration));
        certificate.id = getIdFromLogs(logs);
    }
    if (configuration.logger) {
        configuration.logger.info(`Certificate ${certificate.id} created`);
    }
    return certificate.sync();
};
async function claimCertificates(certificateIds, configuration) {
    const registry = configuration.blockchainProperties.certificateLogicInstance;
    const certificateIdsAsNumber = certificateIds.map(c => parseInt(c, 10));
    const certificatesPromises = certificateIds.map(certId => new Entity(certId, configuration).sync());
    const certificates = await Promise.all(certificatesPromises);
    const isOwnerPromise = certificates.map(cert => cert.isOwned());
    const owned = await Promise.all(isOwnerPromise);
    const ownsAllCertificates = owned.every(isOwned => isOwned === true);
    if (!ownsAllCertificates) {
        throw new Error(`You can only claim your own certificates`);
    }
    const valuesPromise = certificates.map(cert => cert.ownedVolume());
    const values = await Promise.all(valuesPromise);
    const { randomHex, hexToBytes } = configuration.blockchainProperties.web3.utils;
    // TO-DO: replace with proper claim data
    const claimData = certificates.map(cert => hexToBytes(randomHex(32)));
    const data = hexToBytes(randomHex(32));
    const { from } = getAccountFromConfiguration(configuration);
    return registry.safeBatchTransferAndClaimFrom(from, from, certificateIdsAsNumber, values, data, claimData, getAccountFromConfiguration(configuration));
}
exports.claimCertificates = claimCertificates;
async function transferCertificates(certificateIds, to, configuration) {
    const registry = configuration.blockchainProperties.certificateLogicInstance;
    const certificateIdsAsNumber = certificateIds.map(c => parseInt(c, 10));
    const certificatesPromises = certificateIds.map(certId => new Entity(certId, configuration).sync());
    const certificates = await Promise.all(certificatesPromises);
    const valuesPromise = certificates.map(cert => cert.ownedVolume());
    const values = await Promise.all(valuesPromise);
    const { randomHex, hexToBytes } = configuration.blockchainProperties.web3.utils;
    // TO-DO: replace with proper data
    const data = hexToBytes(randomHex(32));
    const { from } = getAccountFromConfiguration(configuration);
    return registry.safeBatchTransferFrom(from, to, certificateIdsAsNumber, values, data, getAccountFromConfiguration(configuration));
}
exports.transferCertificates = transferCertificates;
//# sourceMappingURL=Certificate.js.map